/*
 * Copyright (c) 2017 - 2020, GÉANT
 *
 * Licensed under the Apache License, Version 2.0 (the “License”); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an “AS IS” BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.geant.idpextension.oidc.attribute.transcoding.impl;

import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import org.geant.idpextension.oidc.attribute.transcoding.AbstractOIDCAttributeTranscoder;
import org.opensaml.profile.context.ProfileRequestContext;

import net.minidev.json.JSONArray;
import net.shibboleth.idp.attribute.AttributeDecodingException;
import net.shibboleth.idp.attribute.AttributeEncodingException;
import net.shibboleth.idp.attribute.IdPAttribute;
import net.shibboleth.idp.attribute.IdPAttributeValue;
import net.shibboleth.idp.attribute.ScopedStringAttributeValue;
import net.shibboleth.idp.attribute.transcoding.TranscodingRule;
import net.shibboleth.utilities.java.support.annotation.constraint.NonnullElements;
import net.shibboleth.utilities.java.support.annotation.constraint.NotEmpty;

/**
 * Class encoding scoped string attributes to string json object. Name of the attribute will be set as the key. The
 * string contains attribute value, delimiter and scope catenated. If there are several attribute values they are
 * delimited with delimiter(space is default) or placed to array.
 */
public class OIDCScopedStringAttributeTranscoder extends AbstractOIDCAttributeTranscoder {

    /** Scope delimiter for combining the data. */
    @Nonnull @NotEmpty public static final String PROP_SCOPE_DELIMITER = "oidc.scopeDelimiter";

    /** {@inheritDoc} */
    @Override
    @Nullable protected Object encodeValues(@Nullable final ProfileRequestContext profileRequestContext,
            @Nonnull final IdPAttribute attribute, @Nonnull final TranscodingRule rule)
                    throws AttributeEncodingException {
        
        if (attribute.getValues().isEmpty()) {
            return null;
        }
    
        final String scopeDelimiter = rule.getOrDefault(PROP_SCOPE_DELIMITER, String.class, "@");
        
        final List<String> values = attribute.getValues()
                .stream()
                .filter(ScopedStringAttributeValue.class::isInstance)
                .map(ScopedStringAttributeValue.class::cast)
                .map(v -> v.getValue() + scopeDelimiter + v.getScope())
                .collect(Collectors.toUnmodifiableList());
        
        // Values of arbitrary mapped type placed into JSONArray.
        if (rule.getOrDefault(PROP_ASARRAY, Boolean.class, false)) {
            final JSONArray array = new JSONArray();
            array.addAll(values);
            return array;
        }
        
        // String concatenation.
        
        final String delimiter = rule.getOrDefault(PROP_STRING_DELIMITER, String.class, " ");
        
        boolean stringEmpty = true;
        final StringBuilder attributeString = new StringBuilder();
        for (final String value : values) {
            if (!stringEmpty) {
                attributeString.append(delimiter);
            } else {
                stringEmpty = false;
            }
            attributeString.append(value);
        }
        
        return attributeString.length() != 0 ? attributeString.toString() : null;
    }

    /** {@inheritDoc} */
    @Override
    @Nonnull @NonnullElements protected List<IdPAttributeValue> decodeValues(
            @Nullable final ProfileRequestContext profileRequestContext, @Nonnull final Object input,
            @Nonnull final TranscodingRule rule) throws AttributeDecodingException {
        
        // TODO: implement value decoding
        return Collections.emptyList();
    }

}