/*
 * Copyright (c) 2017 - 2020, GÉANT
 *
 * Licensed under the Apache License, Version 2.0 (the “License”); you may not
 * use this file except in compliance with the License. You may obtain a copy
 * of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an “AS IS” BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.geant.idpextension.oidc.config;

import java.security.Principal;
import java.time.Duration;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.function.Function;
import java.util.function.Predicate;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import org.geant.idpextension.oidc.authn.principal.AuthenticationContextClassReferencePrincipal;
import org.opensaml.profile.context.ProfileRequestContext;

import com.google.common.base.Predicates;

import net.shibboleth.idp.authn.config.AuthenticationProfileConfiguration;
import net.shibboleth.utilities.java.support.annotation.constraint.NonNegative;
import net.shibboleth.utilities.java.support.annotation.constraint.NonnullElements;
import net.shibboleth.utilities.java.support.annotation.constraint.NotEmpty;
import net.shibboleth.utilities.java.support.annotation.constraint.NotLive;
import net.shibboleth.utilities.java.support.annotation.constraint.Positive;
import net.shibboleth.utilities.java.support.annotation.constraint.Unmodifiable;
import net.shibboleth.utilities.java.support.logic.Constraint;
import net.shibboleth.utilities.java.support.logic.FunctionSupport;
import net.shibboleth.utilities.java.support.primitive.StringSupport;

/**
 * Profile configuration for the OpenID Connect authorization and token end points.
 */
public class OIDCCoreProtocolConfiguration extends AbstractOIDCFlowAwareProfileConfiguration
        implements AuthenticationProfileConfiguration {

    /** OIDC base protocol URI. */
    @Nonnull @NotEmpty public static final String PROTOCOL_URI = "http://openid.net/specs/openid-connect-core-1_0.html";

    /** ID for this profile configuration. */
    @Nonnull @NotEmpty public static final String PROFILE_ID = "http://csc.fi/ns/profiles/oidc/sso/browser";

    /** Whether attributes should be resolved in the course of the profile. */
    @Nonnull private Predicate<ProfileRequestContext> resolveAttributesPredicate;

    /** Whether to mandate forced authentication for the request. */
    @Nonnull private Predicate<ProfileRequestContext> forceAuthnPredicate;
    
    /** Whether all acr claim requests should be treated as Essential. */
    @Nonnull private Predicate<ProfileRequestContext> acrRequestAlwaysEssentialPredicate;

    /** Whether client is required to use PKCE. */
    @Nonnull private Predicate<ProfileRequestContext> forcePKCEPredicate;

    /** Whether client is allowed to use PKCE code challenge method plain. */
    @Nonnull private Predicate<ProfileRequestContext> allowPKCEPlainPredicate;

    /** Lookup function to supply proxyCount property. */
    @Nonnull private Function<ProfileRequestContext,Integer> proxyCountLookupStrategy;

    /** Lookup function to supply default authentication methods. */
    @Nonnull private Function<ProfileRequestContext,Collection<AuthenticationContextClassReferencePrincipal>>
            defaultAuthenticationContextsLookupStrategy;

    /** Lookup function to supply authentication flows. */
    @Nonnull private Function<ProfileRequestContext,Set<String>> authenticationFlowsLookupStrategy;

    /** Lookup function to supply post authentication flows. */
    @Nonnull private Function<ProfileRequestContext,Collection<String>> postAuthenticationFlowsLookupStrategy;

    /** Lookup function to supply ID token lifetime. */
    @Nonnull private Function<ProfileRequestContext,Duration> idTokenLifetimeLookupStrategy;

    /** Lookup function to supply lifetime of authz code. */
    @Nonnull private Function<ProfileRequestContext,Duration> authorizeCodeLifetimeLookupStrategy;

    /** Lookup function to supply access token lifetime. */
    @Nonnull private Function<ProfileRequestContext,Duration> accessTokenLifetimeLookupStrategy;

    /** Lookup function to supply refresh token lifetime. */
    @Nonnull private Function<ProfileRequestContext,Duration> refreshTokenLifetimeLookupStrategy;

    /** Lookup function to supply additional audiences for ID token. */
    @Nonnull private Function<ProfileRequestContext,Set<String>> assertionAudiencesLookupStrategy;

    /**
     * Constructor.
     */
    public OIDCCoreProtocolConfiguration() {
        this(PROFILE_ID);
    }

    /**
     * Creates a new configuration instance.
     *
     * @param profileId Unique profile identifier.
     */
    public OIDCCoreProtocolConfiguration(@Nonnull @NotEmpty final String profileId) {
        super(profileId);
        forceAuthnPredicate = Predicates.alwaysFalse();
        resolveAttributesPredicate = Predicates.alwaysTrue();
        proxyCountLookupStrategy = FunctionSupport.constant(null);
        assertionAudiencesLookupStrategy = FunctionSupport.constant(null);
        
        idTokenLifetimeLookupStrategy = FunctionSupport.constant(Duration.ofHours(1));
        authorizeCodeLifetimeLookupStrategy = FunctionSupport.constant(Duration.ofMinutes(5));
        accessTokenLifetimeLookupStrategy = FunctionSupport.constant(Duration.ofMinutes(10));
        refreshTokenLifetimeLookupStrategy = FunctionSupport.constant(Duration.ofHours(2));
        
        acrRequestAlwaysEssentialPredicate = Predicates.alwaysFalse();
        forcePKCEPredicate = Predicates.alwaysFalse();
        allowPKCEPlainPredicate = Predicates.alwaysFalse();
        
        defaultAuthenticationContextsLookupStrategy = FunctionSupport.constant(null);
        authenticationFlowsLookupStrategy = FunctionSupport.constant(null);
        postAuthenticationFlowsLookupStrategy = FunctionSupport.constant(null);
    }
    
    /**
     * Get whether attributes should be resolved during the profile.
     *
     * <p>Default is true</p>
     * 
     * @param profileRequestContext current profile request context
     * 
     * @return true iff attributes should be resolved
     */
    public boolean isResolveAttributes(@Nullable final ProfileRequestContext profileRequestContext) {
        return resolveAttributesPredicate.test(profileRequestContext);
    }

    /**
     * Set whether attributes should be resolved during the profile.
     * 
     * @param flag flag to set
     */
    public void setResolveAttributes(final boolean flag) {
        resolveAttributesPredicate = flag ? Predicates.alwaysTrue() : Predicates.alwaysFalse();
    }
    
    /**
     * Set a condition to determine whether attributes should be resolved during the profile.
     * 
     * @param condition condition to set
     */
    public void setResolveAttributesPredicate(@Nonnull final Predicate<ProfileRequestContext> condition) {
        resolveAttributesPredicate = Constraint.isNotNull(condition, "Resolve attributes predicate cannot be null");
    }
    
    /** {@inheritDoc} */
    public boolean isForceAuthn(@Nullable final ProfileRequestContext profileRequestContext) {
        return forceAuthnPredicate.test(profileRequestContext);
    }
    
    /**
     * Set whether a fresh user presence proof should be required for this request.
     * 
     * @param flag flag to set
     */
    public void setForceAuthn(final boolean flag) {
        forceAuthnPredicate = flag ? Predicates.alwaysTrue() : Predicates.alwaysFalse();
    }
    
    /**
     * Set a condition to determine whether a fresh user presence proof should be required for this request.
     * 
     * @param condition condition to set
     */
    public void setForceAuthnPredicate(@Nonnull final Predicate<ProfileRequestContext> condition) {
        forceAuthnPredicate = Constraint.isNotNull(condition, "Forced authentication predicate cannot be null");
    }
    
    /** {@inheritDoc} */
    @Nullable public Integer getProxyCount(@Nullable final ProfileRequestContext profileRequestContext) {
        final Integer count = proxyCountLookupStrategy.apply(profileRequestContext);
        if (count != null) {
            Constraint.isGreaterThanOrEqual(0, count, "Proxy count must be greater than or equal to 0");
        }
        return count;
    }

    /**
     * Sets the maximum number of times an assertion may be proxied outbound and/or
     * the maximum number of hops between the relying party and a proxied authentication
     * authority inbound.
     * 
     * @param count proxy count
     */
    public void setProxyCount(@Nullable @NonNegative final Integer count) {
        if (count != null) {
            Constraint.isGreaterThanOrEqual(0, count, "Proxy count must be greater than or equal to 0");
        }
        proxyCountLookupStrategy = FunctionSupport.constant(count);
    }

    /**
     * Set a lookup strategy for the maximum number of times an assertion may be proxied outbound and/or
     * the maximum number of hops between the relying party and a proxied authentication authority inbound.
     *
     * @param strategy  lookup strategy
     */
    public void setProxyCountLookupStrategy(@Nonnull final Function<ProfileRequestContext,Integer> strategy) {
        proxyCountLookupStrategy = Constraint.isNotNull(strategy, "Lookup strategy cannot be null");
    }
    
    /** {@inheritDoc} */
    @Nonnull @NonnullElements @NotLive @Unmodifiable public Set<String> getAuthenticationFlows(
            @Nullable final ProfileRequestContext profileRequestContext) {
        final Set<String> flows = authenticationFlowsLookupStrategy.apply(profileRequestContext);
        if (flows != null) {
            return Set.copyOf(flows);
        }
        return Collections.emptySet();
    }

    /**
     * Set the authentication flows to use.
     * 
     * @param flows   flow identifiers to use
     */
    public void setAuthenticationFlows(@Nullable @NonnullElements final Collection<String> flows) {
        if (flows != null) {
            authenticationFlowsLookupStrategy =
                    FunctionSupport.constant(Set.copyOf(StringSupport.normalizeStringCollection(flows)));
        } else {
            authenticationFlowsLookupStrategy = FunctionSupport.constant(null);
        }
    }

    /**
     * Set a lookup strategy for the authentication flows to use.
     *
     * @param strategy  lookup strategy
     */
    public void setAuthenticationFlowsLookupStrategy(
            @Nonnull final Function<ProfileRequestContext,Set<String>> strategy) {
        authenticationFlowsLookupStrategy = Constraint.isNotNull(strategy, "Lookup strategy cannot be null");
    }

    /** {@inheritDoc} */
    @Nonnull @NonnullElements @NotLive @Unmodifiable public List<String> getPostAuthenticationFlows(
            @Nullable final ProfileRequestContext profileRequestContext) {
        final Collection<String> flows = postAuthenticationFlowsLookupStrategy.apply(profileRequestContext);
        if (flows != null) {
            return List.copyOf(flows);
        }
        return Collections.emptyList();
    }

    /**
     * Set the ordered collection of post-authentication interceptor flows to enable.
     * 
     * @param flows   flow identifiers to enable
     */
    public void setPostAuthenticationFlows(@Nullable @NonnullElements final Collection<String> flows) {
        if (flows != null) {
            postAuthenticationFlowsLookupStrategy =
                    FunctionSupport.constant(List.copyOf(StringSupport.normalizeStringCollection(flows)));
        } else {
            postAuthenticationFlowsLookupStrategy = FunctionSupport.constant(null);
        }
    }

    /**
     * Set a lookup strategy for the post-authentication interceptor flows to enable.
     *
     * @param strategy  lookup strategy
     */
    public void setPostAuthenticationFlowsLookupStrategy(
            @Nonnull final Function<ProfileRequestContext,Collection<String>> strategy) {
        postAuthenticationFlowsLookupStrategy = Constraint.isNotNull(strategy, "Lookup strategy cannot be null");
    }

    /** {@inheritDoc} */
    @Nonnull @NonnullElements @NotLive @Unmodifiable public List<Principal> getDefaultAuthenticationMethods(
            @Nullable final ProfileRequestContext profileRequestContext) {
        final Collection<AuthenticationContextClassReferencePrincipal> methods =
                defaultAuthenticationContextsLookupStrategy.apply(profileRequestContext);
        if (methods != null) {
            return List.copyOf(methods);
        }
        return Collections.emptyList();
    }
        
    /**
     * Set the default authentication contexts to use, expressed as custom principals.
     * 
     * @param contexts default authentication contexts to use
     */
    public void setDefaultAuthenticationMethods(
            @Nullable @NonnullElements final Collection<AuthenticationContextClassReferencePrincipal> contexts) {
        if (contexts != null) {
            defaultAuthenticationContextsLookupStrategy = FunctionSupport.constant(List.copyOf(contexts));
        } else {
            defaultAuthenticationContextsLookupStrategy = FunctionSupport.constant(null);
        }
    }

    /**
     * Set a lookup strategy for the authentication contexts to use, expressed as custom principals.
     *
     * @param strategy  lookup strategy
     */
    public void setDefaultAuthenticationMethodsLookupStrategy(
            @Nonnull final Function<ProfileRequestContext,Collection<AuthenticationContextClassReferencePrincipal>>
            strategy) {
        defaultAuthenticationContextsLookupStrategy = Constraint.isNotNull(strategy, "Lookup strategy cannot be null");
    }

    /**
     * Get ID token lifetime.
     * 
     * <p>Defaults to 1 hour.</p>
     * 
     * @param profileRequestContext profile request context
     * 
     * @return ID token lifetime
     */
    @Positive @Nonnull public Duration getIDTokenLifetime(@Nullable final ProfileRequestContext profileRequestContext) {
        
        final Duration lifetime = idTokenLifetimeLookupStrategy.apply(profileRequestContext);
        
        Constraint.isTrue(lifetime != null && !lifetime.isZero() && !lifetime.isNegative(),
                "ID token lifetime must be greater than 0");
        return lifetime;
    }

    /**
     * Set the lifetime of an id token.
     * 
     * @param lifetime lifetime of an id token in milliseconds
     */
    public void setIDTokenLifetime(@Positive @Nonnull final Duration lifetime) {
        Constraint.isTrue(lifetime != null && !lifetime.isZero() && !lifetime.isNegative(),
                "ID token lifetime must be greater than 0");
        
        idTokenLifetimeLookupStrategy = FunctionSupport.constant(lifetime);
    }

    /**
     * Set a lookup strategy for the ID token lifetime.
     *
     * @param strategy lookup strategy
     */
    public void setIDTokenLifetimeLookupStrategy(@Nonnull final Function<ProfileRequestContext,Duration> strategy) {
        idTokenLifetimeLookupStrategy = Constraint.isNotNull(strategy, "Lookup strategy cannot be null");
    }

    /**
     * Get access token lifetime.
     * 
     * <p>Defaults to 10 minutes.</p>
     * 
     * @param profileRequestContext profile request context
     * 
     * @return access token lifetime
     */
    @Positive @Nonnull
    public Duration getAccessTokenLifetime(@Nullable final ProfileRequestContext profileRequestContext) {
        final Duration lifetime = accessTokenLifetimeLookupStrategy.apply(profileRequestContext);
        
        Constraint.isTrue(lifetime != null && !lifetime.isZero() && !lifetime.isNegative(),
                "Access token lifetime must be greater than 0");
        return lifetime;
    }
    
    /**
     * Set the lifetime of an access token.
     * 
     * @param lifetime lifetime of an access token in milliseconds
     */
    public void setAccessTokenLifetime(@Positive @Nonnull final Duration lifetime) {
        Constraint.isTrue(lifetime != null && !lifetime.isZero() && !lifetime.isNegative(),
                "Access token lifetime must be greater than 0");
        
        accessTokenLifetimeLookupStrategy = FunctionSupport.constant(lifetime);
    }

    
    /**
     * Set a lookup strategy for the access token lifetime.
     *
     * @param strategy lookup strategy
     */
    public void setAccessTokenLifetimeLookupStrategy(
            @Nullable final Function<ProfileRequestContext,Duration> strategy) {
        accessTokenLifetimeLookupStrategy = Constraint.isNotNull(strategy, "Lookup strategy cannot be null");
    }

    /**
     * Get authz code lifetime.
     * 
     * <p>Defaults to 5 minutes.</p>
     * 
     * @param profileRequestContext profile request context
     * 
     * @return authz code lifetime
     */
    @Positive @Nonnull
    public Duration getAuthorizeCodeLifetime(@Nullable final ProfileRequestContext profileRequestContext) {
        final Duration lifetime = authorizeCodeLifetimeLookupStrategy.apply(profileRequestContext);
        
        Constraint.isTrue(lifetime != null && !lifetime.isZero() && !lifetime.isNegative(),
                "Authorization code lifetime must be greater than 0");
        return lifetime;
    }

    /**
     * Set the lifetime of authz code.
     * 
     * @param lifetime lifetime of authz code
     */
    public void setAuthorizeCodeLifetime(@Positive @Nonnull final Duration lifetime) {
        Constraint.isTrue(lifetime != null && !lifetime.isZero() && !lifetime.isNegative(),
                "Authorization code lifetime must be greater than 0");
        
        authorizeCodeLifetimeLookupStrategy = FunctionSupport.constant(lifetime);
    }
    
    /**
     * Set a lookup strategy for the authz code lifetime.
     *
     * @param strategy lookup strategy
     */
    public void setAuthorizeCodeLifetimeLookupStrategy(
            @Nonnull final Function<ProfileRequestContext,Duration> strategy) {
        authorizeCodeLifetimeLookupStrategy = Constraint.isNotNull(strategy, "Lookup strategy cannot be null");
    }

    /**
     * Get refresh token lifetime.
     * 
     * <p>Defaults to 2 hours.</p>
     * 
     * @param profileRequestContext profile request context
     * 
     * @return refresh token lifetime
     */
    @Nonnull @Positive
    public Duration getRefreshTokenLifetime(@Nullable final ProfileRequestContext profileRequestContext) {
        
        final Duration lifetime = refreshTokenLifetimeLookupStrategy.apply(profileRequestContext);
        
        Constraint.isTrue(lifetime != null && !lifetime.isZero() && !lifetime.isNegative(),
                "Refresh token lifetime must be greater than 0");
        return lifetime;
    }

    /**
     * Set the lifetime of refresh token.
     * 
     * @param lifetime lifetime of an refresh token
     */
    public void setRefreshTokenLifetime(@Nonnull @Positive final Duration lifetime) {
        Constraint.isTrue(lifetime != null && !lifetime.isZero() && !lifetime.isNegative(),
                "Refresh token lifetime must be greater than 0");
        
        refreshTokenLifetimeLookupStrategy = FunctionSupport.constant(lifetime);
    }

    /**
     * Set a lookup strategy for the refresh token lifetime.
     *
     * @param strategy lookup strategy
     */
    public void setRefreshTokenLifetimeLookupStrategy(
            @Nullable final Function<ProfileRequestContext,Duration> strategy) {
        refreshTokenLifetimeLookupStrategy = Constraint.isNotNull(strategy, "Lookup strategy cannot be null");
    }
    
    /**
     * Get the set of audiences, in addition to the relying party(ies) to which the IdP is issuing the ID Token, with
     * which the token may be shared.
     * 
     * @param profileRequestContext profile request context
     * 
     * @return the additional audiences
     */
    @Nonnull @NonnullElements @NotLive public Set<String> getAdditionalAudiencesForIdToken(
            @Nullable final ProfileRequestContext profileRequestContext) {
        
        final Set<String> audiences = assertionAudiencesLookupStrategy.apply(profileRequestContext);
        if (audiences != null) {
            return Set.copyOf(audiences);
        }
        return Collections.emptySet();
    }

    /**
     * Set the set of audiences, in addition to the relying party(ies) to which the IdP is issuing the ID Token, with
     * which the token may be shared.
     * 
     * @param audiences the additional audiences
     */
    public void setAdditionalAudiencesForIdToken(@Nullable @NonnullElements final Collection<String> audiences) {

        if (audiences == null || audiences.isEmpty()) {
            assertionAudiencesLookupStrategy = FunctionSupport.constant(null);
        } else {
            assertionAudiencesLookupStrategy = FunctionSupport.constant(
                    Set.copyOf(StringSupport.normalizeStringCollection(audiences)));
        }
    }

    /**
     * Set a lookup strategy for the set of audiences, in addition to the relying party(ies) to which the IdP
     * is issuing the ID Token, with which the token may be shared.
     *
     * @param strategy  lookup strategy
     */
    public void setAdditionalAudiencesForIdTokenLookupStrategy(
            @Nonnull final Function<ProfileRequestContext,Set<String>> strategy) {
        assertionAudiencesLookupStrategy = Constraint.isNotNull(strategy, "Lookup strategy cannot be null");
    }

    /**
     * Get whether all acr claim requests should be treated as Essential.
     * 
     * @param profileRequestContext profile request context
     * 
     * @return whether all acr claim requests should be treated as Essential
     */
    public boolean isAcrRequestAlwaysEssential(@Nullable final ProfileRequestContext profileRequestContext) {
        return acrRequestAlwaysEssentialPredicate.test(profileRequestContext);
    }

    /**
     * Set whether all acr claim requests should be treated as Essential.
     * 
     * @param flag flag to set
     */
    public void setAcrRequestAlwaysEssential(final boolean flag) {
        acrRequestAlwaysEssentialPredicate = flag ? Predicates.alwaysTrue() : Predicates.alwaysFalse();
    }
    
    /**
     * Set condition for whether all acr claim requests should be treated as Essential.
     * 
     * @param condition condition to set
     */
    public void setAcrRequestAlwaysEssentialPredicate(@Nonnull final Predicate<ProfileRequestContext> condition) {
        acrRequestAlwaysEssentialPredicate = Constraint.isNotNull(condition, "Condition cannot be null");
    }

    /**
     * Get whether client is required to use PKCE.
     * 
     * @param profileRequestContext profile request context
     * 
     * @return whether client is required to use PKCE
     */
    public boolean isForcePKCE(@Nullable final ProfileRequestContext profileRequestContext) {
        return forcePKCEPredicate.test(profileRequestContext);
    }

    /**
     * Set whether client is required to use PKCE.
     * 
     * @param flag flag to set
     */
    public void setForcePKCE(final boolean flag) {
        forcePKCEPredicate = flag ? Predicates.alwaysTrue() : Predicates.alwaysFalse();
    }

    /**
     * Set condition for whether client is required to use PKCE.
     * 
     * @param condition condition to set
     */
    public void setForcePKCEPredicate(@Nonnull final Predicate<ProfileRequestContext> condition) {
        forcePKCEPredicate = Constraint.isNotNull(condition, "Condition cannot be null");
    }

    /**
     * Get whether client is allowed to use PKCE code challenge method plain.
     * 
     * @param profileRequestContext profile request context
     * 
     * @return whether client is allowed to use PKCE code challenge method plain
     */
    public boolean isAllowPKCEPlain(@Nullable final ProfileRequestContext profileRequestContext) {
        return allowPKCEPlainPredicate.test(profileRequestContext);
    }

    /**
     * Set whether client is allowed to use PKCE code challenge method plain.
     * 
     * @param flag flag to set
     */
    public void setAllowPKCEPlain(final boolean flag) {
        allowPKCEPlainPredicate = flag ? Predicates.alwaysTrue() : Predicates.alwaysFalse();
    }

    /**
     * Set condition for whether client is allowed to use PKCE code challenge method plain.
     * 
     * @param condition condition to set
     */
    public void setAllowPKCEPlain(@Nonnull final Predicate<ProfileRequestContext> condition) {
        allowPKCEPlainPredicate = Constraint.isNotNull(condition, "Condition cannot be null");
    }

}